clc; clear;

% Black Scholes call price
BSC = @(s,K,rf,q,v,T) (s*exp(-q*T)*normcdf((log(s/K) + (rf-q+v^2/2)*T)/v/sqrt(T)) - K*exp(-rf*T)*normcdf((log(s/K) + (rf-q+v^2/2)*T)/v/sqrt(T) - v*sqrt(T)));

% Load the DJIA data
IV = [...
    19.62	19.47	20.19	21.15;
    19.10	19.05	19.80	20.82;
    18.60	18.61	19.43	20.57;
    18.10	18.12	19.07	20.21;
    17.61	17.74	18.71	20.00;
    17.18	17.43	18.42	19.74;
    16.71	17.06	18.13	19.50;
    16.44	16.71	17.83	19.27;
    16.45	16.41	17.60	18.99;
    16.61	16.25	17.43	18.84;
    17.01	16.02	17.26	18.62;
    17.55	16.10	17.16	18.46;
    17.96	16.57	17.24	18.42]./100;
K = (124:136);
T = [37 72 135 226]./365;
[NK NT] = size(IV);
S0 = 129.14;
r = 0.0010;
q = 0.0068;

% Perturb the IV to introduce arbitrage
a = -0.01;
b = +0.01;
randIV = a + (b-a).*rand(NK,NT);
IV = IV + randIV;

% Market prices of calls with arbitrage in the IV
for k=1:NK
    for t=1:NT
        MktPrice(k,t) = BSC(S0,K(k),r,q,IV(k,t),T(t));
    end
end

% x-values are the strikes
x = K;
n = length(x);

% x-Values at which to fit the splines
xi = [min(x):.5:max(x)];
Nxi = length(xi);

% Smoothing parameter (0 = smoothing)
lambda = 0.05;

% Get the regular smoothed spline at each maturity
arbitrage = 'N';
for t=1:NT
    % Observed price
    y = MktPrice(:,t)';
    % Run the smoothed spline, allowing for arbitrage
    [greg(:,t) gamreg(:,t)] = NSplineFengler(x,y,lambda,arbitrage,[],[],[],[]);
    % Obtain the splined prices.
    Yf(:,t) = FitNaturalSpline(xi,x,gamreg(:,t),greg(:,t));
end

% Prices obtained with spline without arbitrage-free consideration
subplot(2,2,1)
plot(x,MktPrice(:,1),'ko',xi,Yf(:,1),'r-',...
     x,MktPrice(:,2),'ko',xi,Yf(:,2),'r-',...
     x,MktPrice(:,3),'ko',xi,Yf(:,3),'r-',...
     x,MktPrice(:,4),'ko',xi,Yf(:,4),'r-');
title('Call prices with smoothing, arbitrage') 
 
%% Fengler algorithm for arbitrage-free IVS.

for t = NT:-1:1;
    % Maturity and prices
    tau = T(t);
    y = MktPrice(:,t)';
    
    % Constraints on g(2)-g(1) and on g(n-1)-g(n)
    bineq = [exp(-r*tau)*(K(2)-K(1)); 0];
    Aineq = [1 -1 zeros(1,2*n-4); zeros(1,n-2) -1 1 zeros(1,n-2)];

    % Lower bound on g(1),g(2),...,g(n),gam(1),gam(2),..,gam(n-2)
    lb = [-exp(-q*tau)*S0*K(1) zeros(1,2*n-3)];

    % Upper bound is different, depending on the maturity
    if t==NT
        % Bound on g(1) only
        ub = [exp(-q*tau)*S0 inf(1,2*n-3)];
    else
        % Bound on g(1),...,g(n) based on g(i) from regular splines at t+1
        ub = [exp(q*(T(t+1)-T(t))).*greg(:,t+1)' inf(1,n-2)];
    end;
    
    % Obtain the no-arbitrage vectors "g" and "gamma"
    arbitrage = 'Y';
    [g(:,t) gam(:,t)] = NSplineFengler(x,y,lambda,arbitrage,Aineq,bineq,lb,ub);
end

% Fit the arbitrage-free 
for t=1:NT
    Yf2(:,t) = FitNaturalSpline(xi,x,gam(:,t),g(:,t));
end

% Spline with no arbitrage-free consideration
subplot(2,2,2)
plot(x,MktPrice(:,1),'ko',xi,Yf2(:,1),'r-',...
     x,MktPrice(:,2),'ko',xi,Yf2(:,2),'r-',...
     x,MktPrice(:,3),'ko',xi,Yf2(:,3),'r-',...
     x,MktPrice(:,4),'ko',xi,Yf2(:,4),'r-');
title('Call prices with smoothing, no arbitrage') 

%% Verify no-arbitrage against maturity by plotting total volatility
a = 0.01;
b = 2;
Tol = 1e-6;
MaxIter = 1e+6;

for t=1:NT
    % Total variance
    for k=1:Nxi
        IVnoarb = BisecBSIV('C',S0,xi(k),r,q,T(t),a,b,Yf2(k,t),Tol,MaxIter);
        TotalVar(k,t) = T(t)*IVnoarb^2;
    end
    % Risk neutral density
    [RND(:,t) K2(:,t)] = ExtractRND(xi,Yf2(:,t));
end

% Plot the total variance
subplot(2,2,3)
plot(xi,TotalVar(:,1),xi,TotalVar(:,2),xi,TotalVar(:,3),xi,TotalVar(:,4))
title('Total var. with smoothing,no arbitrage') 

% Plot the risk neutral densities
subplot(2,2,4)
plot(K2(:,1),RND(:,1),K2(:,2),RND(:,2),K2(:,3),RND(:,3),K2(:,4),RND(:,4))
title('RND with smoothing,no arbitrage') 

