function [g gam] = NSplineFengler(x,y,lambda,NoArbitrage,Aineq,bineq,lb,ub)

% Natural spline using Reinsch algorithm described in
% Mathias Fengler
% Arbitrage-free smoothing of the implied volatility surface

% INPUTS
%   x = vector of x-values
%   y = vector of function values evaluated at x
%   lambda = smoothing parameter >= 0
%   NoArbitrage = 'N' (arbitrage allowed) 'Y' (arbitrage not allowed)
%   Aineq = "A" matrix for the inequality Ax <= b in fmincon
%   bineq = "b" vector for the inequality Ax <= b in fmincon
%   lb = lower bound on g vector
%   ub = upper bound on g vector
% OUTPUTS
%   g = vector g consists of spline knots
%   gam = gamma vector consists of derivatives

% length of inputs
n = length(x);

% Create the h vector
for i=1:n-1
    h(i) = x(i+1)-x(i);
end

% Create the R matrix
R = zeros(n-2,n-2);
for i=1:n-2;
    R(i,i) = (h(i)+h(i+1))/3;
    if i<=n-3
        R(i,i+1) = h(i+1)/6;
    end
    if i>=2
        R(i,i-1) = h(i)/6;
    end
end

% Create the Q matrix
Q = zeros(n,n-2);
for i=1:n-2
    Q(i,i) = 1/h(i);
end
for i=3:n
    Q(i,i-2) = 1/h(+1);
end
for i=2:n-1
    Q(i,i-1) = -1/h(i-1) - 1/h(i);
end

% Create the A and B matrices
A = [Q; -(R)];
B = [eye(n) zeros(n,n-2); zeros(n-2,n) lambda.*R];

% Augment the y vector of functional values
ya = [y zeros(1,n-2)];

% Define the objective function
NSplineObjFun = inline('-y*x + 0.5.*transpose(x)*B*x','x','y','B');

% Define the starting values
start = zeros(2*n-2,1);

% Define the constraint matrix and vector 
Aeq = transpose(A);
beq   = zeros(n-2,1);

% Find the vector X through optimization
options = optimset('Algorithm','active-set');

% Different forms depending on whether no-arbitrage constraints are imposed
if strcmp(NoArbitrage,'N')
    % Uses Ax = b only
    Y = fmincon(@(p) NSplineObjFun(p,ya,B), start,[],[],Aeq,beq,[],[],[],options);
else
    % Uses A*x = b, Aineq*x <= bineq, x<=lb, x>=ub 
    Y = fmincon(@(x) NSplineObjFun(x,ya,B), start,Aineq,bineq,Aeq,beq,lb,ub,[],options);
end

% First n values of X are the points gi
g = Y(1:n);

% Remaining values of X are the derivatives
gam = Y(n+1:2*n-2);

% Augment gamma vector with gam(1) = gam(n) = 0
gam = [0; gam; 0];

